import 'dart:core';

import 'package:json_annotation/json_annotation.dart';
import 'package:lori_sdk/data/models/cart/cart.dart';
import 'package:lori_sdk/data/models/product/featuredCharacteristic.dart';
import 'package:lori_sdk/data/models/product/product.dart';

part 'cartItem.g.dart';

const CART_ITEM_CURRENT_VERSION = 3;

@JsonSerializable()
class CartItem implements Comparable<CartItem> {
  CartItem();

  /// The CartItem unique identifier
  String id;

  /// The sku related to this item
  Sku sku;

  /// The quantity requested by the user for this item
  int quantity;

  /// The quantity available on the seller for this item
  int quantityAvailable;

  /// The quantity separated (a.k.a. that can be bought) for this item
  int quantitySeparated;

  /// Indicates if the cartItem is a regular one or a promotional one
  @JsonKey(defaultValue: false)
  bool isPromotional;

  /// Indicates if the cartItem is a regular one or a gift one
  @JsonKey(defaultValue: false)
  bool isGift;

  @Deprecated("Please use subTotalAmount instead")
  double price;
  @Deprecated("Please use subTotalSeparatedAmount instead")
  double priceSeparated;

  /// The price of each unit of this item, without any modifiers
  double unitPrice;

  /// The discount price applied to this item (if any)
  ///
  /// The discount price will ALWAYS be applied to ONLY ONE unit of the item
  double discountPrice;

  /// The subtotal price of this item, without any modifiers
  ///
  /// (a.k.a. unitPrice * quantity)
  double subTotalAmount;

  /// The subtotal price of this item, considering only the SEPARATED ITEMS and without any modifiers
  ///
  /// (a.k.a. unitPrice * quantitySeparated
  double subTotalSeparatedAmount;

  /// The total price of this item, applying the modifiers
  ///
  /// (a.k.a. subTotalAmount - discountPrice)
  double totalAmount;

  /// The total price of this item, considering only the SEPARATED ITEMS and applying the modifiers
  ///
  /// (a.k.a. subTotalSeparatedAmount - discountPrice)
  double totalSeparatedAmount;

  /// The CartItem has an expiration date
  ///
  /// Sometimes this values might be imported from a previous cart item,
  ///   please take a look at isUpdatingFrom(CartItem cartItem) method.
  ///
  /// Defaults to `this.defaultExpirationDate` when null.
  DateTime expirationDate;

  bool usingWholesalePrice;

  @JsonKey(defaultValue: [])
  List<FeaturedCharacteristic> featuredCharacteristics;

  /// The map containing the changes on the lastUpdate
  ///
  /// This map will be updated at `isUpdatingFrom`, so this means that
  ///   at the second update it will already be lost
  @JsonKey(defaultValue: const {})
  Map<String, dynamic> changesFromLastUpdate;

  void migrateToCurrentVersion() {
    if (CART_ITEM_CURRENT_VERSION > 1) {
      //Create the new price values from "price" and "priceSeparated"
      if (this.unitPrice == null) this.unitPrice = this.price / this.quantity;
      if (this.discountPrice == null) this.discountPrice = 0.0;
      if (this.subTotalAmount == null) this.subTotalAmount = this.price;
      if (this.subTotalSeparatedAmount == null) this.subTotalSeparatedAmount = this.priceSeparated;
      if (this.totalAmount == null) this.totalAmount = this.price;
      if (this.totalSeparatedAmount == null) this.totalSeparatedAmount = this.priceSeparated;
    }
  }

  void isUpdatingFrom(CartItem cartItem, {bool forceUpdate = false}) {
    if (!forceUpdate) {
      // If the quantity of the item has not changed,
      //    this means the item has not been updated
      if (this.quantity == cartItem.quantity) {
        // Since the item has not been updated,
        //    recover the expirationDate from the previous one
        this.expirationDate = cartItem.expirationDate;
      }
    }

    if (this.unitPrice != cartItem.unitPrice) {
      changesFromLastUpdate['unitPrice'] = cartItem.unitPrice;
    }

    if (this.quantity != cartItem.quantity) {
      changesFromLastUpdate['quantity'] = cartItem.quantity;
    }

    if (this.quantity == cartItem.quantity && this.totalAmount != cartItem.totalAmount) {
      changesFromLastUpdate['totalAmount'] = cartItem.totalAmount;
    }
  }

  /// Some shorthand methods to simplify the check of `changesFromLastUpdate`
  ///   for the most used variables
  ///
  bool get hasChanged {
    return changesFromLastUpdate.keys.length > 0;
  }

  bool get hasChangedUnitPrice {
    return changesFromLastUpdate.containsKey('unitPrice');
  }

  bool get hasChangedQuantity {
    return changesFromLastUpdate.containsKey('unitPrice');
  }

  bool get hasChangedTotalAmount {
    return changesFromLastUpdate.containsKey('totalAmount');
  }

  bool get hasChangedPrice {
    return this.hasChangedUnitPrice || this.hasChangedTotalAmount;
  }

  @override
  int compareTo(CartItem other) {
    //return -1 if this is before other
    //return 0 if this equal other
    //return 1 if this is after other

    //First of all we order by the cart Items types
    //First, come the regular items
    //Second, come the promotional items
    //Third, come the gift items

    //If this is not promotional and the other is we return -1
    if (!this.isPromotional && other.isPromotional) return -1;

    //If this is not gift and the other is we return -1
    if (!this.isGift && other.isGift) return -1;

    //After all this, we order them by unit price
    return this.unitPrice.compareTo(other.unitPrice);
  }

  factory CartItem.fromJson(Map<String, dynamic> json) {
    var cartItem = _$CartItemFromJson(json);
    if (cartItem.expirationDate == null) cartItem.expirationDate = Cart.defaultExpirationDate;
    return cartItem;
  }

  Map<String, dynamic> toJson() => _$CartItemToJson(this);
}
